#/**********************************************************************
#   Copyright (C) Christopher Yeoh <cyeoh@samba.org> 2005
#   
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#**********************************************************************/
#include <assert.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "SCClient.H"
#include "protocol.h"

StateStrings = [
    "Dead",
    "Shielded",
    "Amnesia",
    "Amnesia (Permanent)",
    "Confused",
    "Confused (Permanent)",
    "Charmed",
    "Charmed (Permanent)",
    "Paralysed",
    "Paralysed (Permanent)",
    "Afraid",
    "Afraid (Permanent)",
    "Anti Spelled",
    "Protected from Evil",
    "Protected from Evil (Permanent)",
    "Heat Resistant",
    "Cold Resistant",
    "Diseased",
    "Poisoned",
    "Blinded",
    "Blinded (Permanent)",
    "Invisible",
    "Invisible (Permanent)",
    "Hasted",
    "Hasted (Permanent)",
    "TimeStopped",
    "Delayed Effect",
    "Permanency",
    "Surrendered"
]
    

SCClient::SCClient(int Connection, string Name, bool Observer)
    : remoteConnection(Connection), rcvdAllGestures(false),
      rcvdAllSpells(false), rcvdAllCreatureInfo(false),
      rcvdAllMonsterResults(false)
{
    // Initialise connection to server
	readData()
    while (!remoteConnection.haveMessageToRead()) {
	readData();
    }

    // Get protocol version and welcome message 
    Message *msg = remoteConnection.getNextMessage();
    assert(msg->command==MSG_ASK_WELCOME);
    int index = 0;
    memcpy(&protocolVersion, msg->data, sizeof(int));
    index += sizeof(int);
    protocolVersion = ntohl(protocolVersion);
    
    assert(protocolVersion==SPELLCAST_PROTOCOL_VERSION);
    serverWelcomeMessage = msg->data + sizeof(int);
    free(msg);

    // Send client details
    msg = Message::createMessage(MSG_RCV_CLIENT_DETAILS, 
				 Name.length()+1);
    strcpy(msg->data, Name.c_str());
    assert(msg->Ok());
    remoteConnection.writeMessage(msg);
    free(msg);
}

void
SCClient::readData()
{
    remoteConnection.readData();
}

bool
SCClient::messageAvailable() const
{
    return remoteConnection.haveMessageToRead();
}

void
SCClient::writeData()
{
    remoteConnection.writeData();
}

bool
SCClient::haveDataToWrite() const
{
    return remoteConnection.pendingDataToWrite();
}

int
SCClient::getConnectionFD() const
{
    return remoteConnection.getFD();
}

const string &
SCClient::getServerWelcomeMessage() const
{
    return serverWelcomeMessage;
}

int
SCClient::getProtocolVersion() const
{
    return protocolVersion;
}

int
SCClient::getOwnID() const
{
    return ownID;
}

int 
SCClient::getTurnTimeout() const
{
    return turnTimeout;
}

int 
SCClient::getMaxRounds() const
{
    return maxRounds;
}

int
SCClient::processMessage()
{
    assert(remoteConnection.haveMessageToRead());

    Message *msg = remoteConnection.getNextMessage();
    assert(msg->Ok());

    switch (msg->command) {
    case MSG_SEND_NEWPLAYER_INFO:
	rcvNewPlayerInfo(msg);
	break;

    case MSG_SEND_ROUND_BEGIN:
	// Start of round
	rcvdAllGestures = false;
	rcvdAllSpells = false;
	rcvdAllCreatureInfo = false;
	rcvdAllMonsterResults = false;
	spellResults.clear();
	monsterResults.clear();
	break;

    case MSG_ASK_FOR_GESTURES:
	break;

    case MSG_SEND_GESTURES_SEEN:
	rcvGesturesSeen(msg);
	break;
	
    case MSG_ASK_FOR_SPELLS_CAST:
	rcvAskForSpellsCast(msg);
	break;

    case MSG_ASK_SPELL_DIRECTIONS:
	rcvAskSpellDirections(msg);
	break;

    case MSG_SEND_SPELL_CAST:
	rcvSpellCast(msg);
	break;

    case MSG_SEND_CREATURE_STATE:
	rcvCreatureState(msg);
	break;

    case MSG_ASK_MONSTER_DIRECTIONS:
	rcvAskMonsterDirections(msg);
	break;

    case MSG_SEND_NEW_MONSTER_INFO:
	rcvNewMonsterInfo(msg);
	break;

    case MSG_SEND_MONSTER_ATTACK_INFO:
	rcvMonsterAttackInfo(msg);
	break;

    case MSG_SEND_END_GAME:
	rcvEndGame(msg);
	break;

    case MSG_SEND_EVENT_INFO:
	rcvEventMessage(msg);
	break;

    case MSG_SEND_START_GAME:
	rcvStartGame(msg);
	break;

    case MSG_ASK_CHARM_PERSON_CTRL_HAND:
	rcvCharmPersonCtrlHand(msg);
	break;

    case MSG_ASK_PARALYSIS_CTRL_HAND:
	rcvParalysisCtrlHand(msg);
	break;

    case MSG_ASK_CHARM_PERSON_CTRL_GESTURE:
	rcvCharmPersonCtrlGesture(msg);
	break;

    default:
	printf("Unknown message %i received\n", msg->command);


    }
    return msg->command;
}

void
SCClient::sendGestures(Gesture LeftHand, Gesture RightHand)
{
    int left, right;
    Message *msg = Message::createMessage(MSG_RCV_GESTURES_USED,
					  sizeof(int)*2);
    left = htonl(LeftHand);
    right = htonl(RightHand);

    memcpy(msg->data, &left, sizeof(int));
    memcpy(msg->data+sizeof(int), &right, sizeof(int));

    assert(msg->Ok());
    
    remoteConnection.writeMessage(msg);
    free(msg);
}

void
SCClient::getCreatureInfo(int ID, CreatureState *Creature) const
{
    assert(creatureData.find(ID)!=creatureData.end());
    *Creature =  creatureData.find(ID)->second;
}

bool
SCClient::haveAllCreatureInfo() const
{
    return rcvdAllCreatureInfo;
}

void
SCClient::getAllCreatureIDs(vector<int> *CreatureIDs) const
{
    CreatureIDs->clear();
    map<int, CreatureState>::const_iterator creature;

    for (creature=creatureData.begin(); creature!=creatureData.end();
	 creature++) {
	CreatureIDs->push_back(creature->first);
    }
}

bool
SCClient::haveAllGestures() const
{
    return rcvdAllGestures;
}

void
SCClient::getGesturesForPlayer(int ID, Gesture *Left, Gesture *Right,
			       bool *PlayerAntiSpelled)
{
    assert(creatureData.find(ID)!=creatureData.end());
    assert(creatureData.find(ID)->second.owner==0);

    if (currentGestureMapping.find(ID)==currentGestureMapping.end()) {
	// Other player probably disconnected
	// This can happen if the player disconnects before first
	// gestures are sent
	*Left = GST_NOTHING;
	*Right = GST_NOTHING;
	*PlayerAntiSpelled = false;
	
    } else {
	*Left = currentGestureMapping.find(ID)->second.primaryGesture;
	*Right = currentGestureMapping.find(ID)->second.secondaryGesture;
	*PlayerAntiSpelled = currentGestureMappingAntiSpelled.find(ID)->second;
    }
}

void
SCClient::getPossibleSpells(vector<Spell> *LeftHandSpells,
			    vector<Spell> *RightHandSpell) 
{
    *LeftHandSpells = possibleLeftHandSpells;
    *RightHandSpell = possibleRightHandSpells;
}

void
SCClient::sendSpellSelection(Spell LeftHandSpell, Spell RightHandSpell)
{
    int leftSpell, rightSpell;
    Message *msg = Message::createMessage(MSG_RCV_SPELLS_CAST,
					  sizeof(int)*2);
    leftSpell = htonl((int)LeftHandSpell);
    rightSpell = htonl((int)RightHandSpell);
    memcpy(msg->data, &leftSpell, sizeof(leftSpell));
    memcpy(msg->data+sizeof(leftSpell), &rightSpell,
	   sizeof(rightSpell));

    assert(msg->Ok());

    remoteConnection.writeMessage(msg);
    free(msg);
}

void
SCClient::getSpellsCast(Spell *LeftHandSpell, Spell *RightHandSpell,
			vector<int> *ValidTargetIDs)
{
    *LeftHandSpell = leftSpellCast;
    *RightHandSpell = rightSpellCast;
    *ValidTargetIDs = validTargets;
}

const vector<SCClient::SpellResult> &
SCClient::getSpellResults() const
{
    return spellResults;
}

bool
SCClient::haveAllSpellsCast() const
{
    return rcvdAllSpells;
}

void
SCClient::GetCharmedHandWizard(int *ID) const
{
    *ID = charmTargetID;
}

void
SCClient::SendCharmedHandWizard(int ID, int Hand) 
{
    Message *msg = Message::createMessage(MSG_RCV_CHARM_PERSON_CTRL_HAND,
					  sizeof(int)*2);
    ID = htonl(ID);
    Hand = htonl(Hand);
    
    memcpy(msg->data, &ID, sizeof(int));
    memcpy(msg->data+sizeof(int), &Hand, sizeof(int));

    assert(msg->Ok());
    remoteConnection.writeMessage(msg);
    free(msg);
}

void
SCClient::GetParalysedHandWizard(int *ID) const
{
    *ID = paralysisTargetID;
}

void
SCClient::SendParalysedHandWizard(int ID, int Hand)
{
    Message *msg = Message::createMessage(MSG_RCV_PARALYSIS_CTRL_HAND,
					  sizeof(int)*2);
    
    ID = htonl(ID);
    Hand = htonl(Hand);

    memcpy(msg->data, &ID, sizeof(int));
    memcpy(msg->data+sizeof(int), &Hand, sizeof(int));

    assert(msg->Ok());
    remoteConnection.writeMessage(msg);
    free(msg);
}

void
SCClient::GetCharmedWizard(int *ID, int *Hand) const
{
    *ID = charmTargetID;
    *Hand = charmedHand;
}

void
SCClient::SendCharmedWizard(int ID, Gesture NewGesture)
{
    Message *msg = Message::createMessage(MSG_RCV_CHARM_PERSON_CTRL_GESTURE,
					  sizeof(int)*2);
    
    ID = htonl(ID);
    int gesture = htonl((int)NewGesture);

    memcpy(msg->data, &ID, sizeof(int));
    memcpy(msg->data+sizeof(int), &gesture, sizeof(int));

    assert(msg->Ok());
    remoteConnection.writeMessage(msg);
    free(msg);
}

const vector<int> &
SCClient::getMonstersToDirect(vector<int> *ValidTargetIDs) const
{
    *ValidTargetIDs = validTargets;
    return monstersToDirect;
}

void
SCClient::sendMonsterTargets(const map<int, int> &MonsterTargets)
{
    assert(monstersToDirect.size()>0);
    map<int, int>::const_iterator monster;
    int numMonsters = monstersToDirect.size();


    Message *msg = Message::createMessage(MSG_RCV_MONSTER_DIRECTIONS,
					  sizeof(int) + 
					  sizeof(int)*numMonsters*2);

    int index = 0;
    numMonsters = htonl(numMonsters);
    memcpy(msg->data+index, &numMonsters, sizeof(int));
    index += sizeof(int);
    int monsterID;
    int targetID;

    for (monster=MonsterTargets.begin(); monster!=MonsterTargets.end();
	 monster++) {

	monsterID = htonl(monster->first);
	targetID = htonl(monster->second);
	
	memcpy(msg->data+index, &monsterID, sizeof(int));
	index += sizeof(int);
	memcpy(msg->data+index, &targetID, sizeof(int));
	index += sizeof(int);
    }
    
    assert(msg->Ok());
    remoteConnection.writeMessage(msg);
    free(msg);
}

const vector<SCClient::MonsterAttackResult> &
SCClient::getMonsterAttackInfo() const
{
    return monsterResults;
}

bool
SCClient::haveAllMonsterAttackInfo() const
{
    return rcvdAllMonsterResults;
}

void
SCClient::sendSpellTargets(int LeftHandTarget, int RightHandTarget)
{
    Message *msg = Message::createMessage(MSG_RCV_SPELL_DIRECTIONS,
					  sizeof(int)*2);
    LeftHandTarget = htonl(LeftHandTarget);
    RightHandTarget = htonl(RightHandTarget);
    memcpy(msg->data, &LeftHandTarget, sizeof(LeftHandTarget));
    memcpy(msg->data+sizeof(LeftHandTarget), &RightHandTarget,
	   sizeof(RightHandTarget));

    assert(msg->Ok());

    remoteConnection.writeMessage(msg);
    free(msg);
}

const vector<int> &
SCClient::getGameWinner() const
{
    return gameWinners;
}

string
SCClient::decodeMessage(const SpellResult &SR)
{
    string msg = SR.resultMessage;
    CreatureState creature;

    getCreatureInfo(SR.sourcePlayer, &creature);
    string spName = creature.name;

    getCreatureInfo(SR.targetCreature, &creature);
    string tcName = creature.name;

    string::size_type index = msg.find("%S");
    while (index!=string::npos) {
	msg.replace(index, 2, spName);
	index = msg.find("%S");
    }

    index = msg.find("%T");
    while (index!=string::npos) {
	msg.replace(index, 2, tcName);
	index = msg.find("%T");
    }

    index = msg.find("%M");
    string summonCreatureName = "***";
    if (index!=string::npos) {
	summonCreatureName = Spells::getCreatureName(SR.spellCast);
    }
    while (index!=string::npos) {
	msg.replace(index, 2, summonCreatureName);
	index = msg.find("%M");
    }

    return msg;
}

string
SCClient::decodeMessage(const MonsterAttackResult &Mar)
{
    string msg = Mar.resultMessage;
    CreatureState creature;

    getCreatureInfo(Mar.source, &creature);
    string source = creature.name;

    getCreatureInfo(Mar.target, &creature);
    string target = creature.name;

    string::size_type index = msg.find("%S");
    while (index!=string::npos) {
	msg.replace(index, 2, source);
	index = msg.find("%S");
    }

    index = msg.find("%T");
    while (index!=string::npos) {
	msg.replace(index, 2, target);
	index = msg.find("%T");
    }

    char *damageString;
    asprintf(&damageString, "%i", Mar.damage);
    index = msg.find("%D");
    while (index!=string::npos) {
	msg.replace(index, 2, damageString);
	index = msg.find("%D");
    }
    free(damageString);

    return msg;
}


const SCClient::EventMessage &
SCClient::getEventMessage() const
{
    return lastEventMessage;
}

string
SCClient::decodeEventMessage(const EventMessage &EventData) {
    string msg = EventData.MessageString;
    CreatureState creature;

    string source = "Unknown";
    if (EventData.Source) {
	getCreatureInfo(EventData.Source, &creature);
	source = creature.name;
    }

    string target = "Unknown";
    if (EventData.Target) {
	getCreatureInfo(EventData.Target, &creature);
	target = creature.name;
    }

    string::size_type index = msg.find("%S");
    while (index!=string::npos) {
	msg.replace(index, 2, source);
	index = msg.find("%S");
    }

    index = msg.find("%T");
    while (index!=string::npos) {
	msg.replace(index, 2, target);
	index = msg.find("%T");
    }

    index = msg.find("%P");
    string spellName = "***";
    if (index!=string::npos) {
	spellName = Spells::getSpellName((Spell)EventData.Misc);
    }
    while (index!=string::npos) {
	msg.replace(index, 2, spellName);
	index = msg.find("%P");
    }

    return msg;
}

string
SCClient::decodeCreatureState(int CreatureState)
{
    string state;

    for (int i=0; i<30; i++) {
	if ( (CreatureState>>i) & 1) {
	    if (state!="") {
		state += ", ";
	    }
	    state += StateStrings[i];
	}
    }
    return state;
}

void
SCClient::rcvStartGame(Message *Msg)
{
    memcpy(&ownID, Msg->data, sizeof(int));
    ownID = ntohl(ownID);
    
    memcpy(&turnTimeout, Msg->data+sizeof(int), sizeof(int));
    turnTimeout = ntohl(turnTimeout);

    memcpy(&maxRounds, Msg->data+sizeof(int)*2, sizeof(int));
    maxRounds = ntohl(maxRounds);

    // Need to reset any information from previous games  we have as its now 
    // out of date as a new game has started
    creatureData.clear();
    currentGestureMapping.clear();
    currentGestureMappingAntiSpelled.clear();
    validTargets.clear();
    spellResults.clear();
    monsterResults.clear();
    possibleLeftHandSpells.clear();
    possibleRightHandSpells.clear();
    gameWinners.clear();
    rcvdAllGestures = false;
    rcvdAllSpells = false;
    rcvdAllCreatureInfo = false;
    rcvdAllMonsterResults = false;
}

void
SCClient::rcvNewPlayerInfo(Message *Msg)
{
    int ID;
    
    memcpy(&ID, Msg->data, sizeof(int));
    ID = ntohl(ID);

    string name = Msg->data+2*sizeof(int);

    CreatureState cs;
    cs.name = name;
    cs.hitPoints = 0;
    cs.state = 0;
    cs.owner = 0;
    cs.type = 0;
    assert(creatureData.find(ID)==creatureData.end());
    creatureData[ID] = cs;
}

void
SCClient::rcvGesturesSeen(Message *Msg)
{
    int ID, leftGesture, rightGesture, antiSpelled;

    memcpy(&ID, Msg->data, sizeof(int));
    ID = ntohl(ID);

    if (ID==0) {
	// End marker 
	rcvdAllGestures = true;
    } else {
	memcpy(&leftGesture, Msg->data+sizeof(int), sizeof(int));
	memcpy(&rightGesture, Msg->data+sizeof(int)*2, sizeof(int));
	memcpy(&antiSpelled, Msg->data+sizeof(int)*3, sizeof(int));

	leftGesture = ntohl(leftGesture);
	rightGesture = ntohl(rightGesture);
	currentGestureMapping[ID]
	    = Spells::GesturePair(Gesture(leftGesture), Gesture(rightGesture));
	if (antiSpelled) {
	    currentGestureMappingAntiSpelled[ID] = true;
	} else {
	    currentGestureMappingAntiSpelled[ID] = false;
	}
    }
}

void
SCClient::rcvAskSpellDirections(Message *Msg)
{
    int leftSpell, rightSpell, numTargets;
    int index = 0;

    memcpy(&leftSpell, Msg->data + index, sizeof(leftSpell));
    index += sizeof(leftSpell);
    memcpy(&rightSpell, Msg->data + index, sizeof(rightSpell));
    index += sizeof(rightSpell);
    memcpy(&numTargets, Msg->data + index, sizeof(numTargets));
    index += sizeof(numTargets);

    numTargets = ntohl(numTargets);

    int targetID;
    validTargets.clear();
    for (int i=0; i<numTargets; i++) {
	memcpy(&targetID, Msg->data+index, sizeof(targetID));
	index += sizeof(targetID);
	targetID = ntohl(targetID);
	validTargets.push_back(targetID);
    }

    leftSpellCast = Spell(ntohl(leftSpell));
    rightSpellCast = Spell(ntohl(rightSpell));
}

void
SCClient::rcvSpellCast(Message *Msg)
{
    SpellResult sr;
    int index = 0;

    memcpy(&sr.sourcePlayer, Msg->data+index, sizeof(int));
    index += sizeof(int);
    sr.sourcePlayer = ntohl(sr.sourcePlayer);

    if (sr.sourcePlayer==0) {
	// End marker
	rcvdAllSpells = true;
    } else {

	memcpy(&sr.targetCreature, Msg->data+index, sizeof(int));
	index += sizeof(int);
	sr.targetCreature = ntohl(sr.targetCreature);
	
	memcpy(&sr.spellCast, Msg->data+index, sizeof(int));
	index += sizeof(int);
	sr.spellCast = (Spell)ntohl(sr.spellCast);
	
	int worked;
	memcpy(&worked, Msg->data+index, sizeof(worked));
	index += sizeof(int);
	sr.spellWorked = (bool)worked;
	
	// skip length (trust server)
	index += sizeof(int);
	sr.resultMessage = Msg->data+index;

	spellResults.push_back(sr);
    }
}

void 
SCClient::rcvCreatureState(Message *Msg)
{
    int creatureID, hp, state;
    int protfromevil, disease, poison, blindness, invisibility;

    memcpy(&creatureID, Msg->data, sizeof(int));
    creatureID = ntohl(creatureID);

    if (creatureID==0) {
	// End marker
	rcvdAllCreatureInfo = true;
    } else {

	memcpy(&hp, Msg->data+sizeof(int), sizeof(int));
	memcpy(&state, Msg->data+sizeof(int)*2, sizeof(int));
	memcpy(&protfromevil, Msg->data+sizeof(int)*3, sizeof(int));
	memcpy(&disease, Msg->data+sizeof(int)*4, sizeof(int));
	memcpy(&poison, Msg->data+sizeof(int)*5, sizeof(int));
	memcpy(&blindness, Msg->data+sizeof(int)*6, sizeof(int));
	memcpy(&invisibility, Msg->data+sizeof(int)*7, sizeof(int));
	
	assert(creatureData.find(creatureID)!=creatureData.end());
	creatureData[creatureID].hitPoints = ntohl(hp);
	creatureData[creatureID].state = ntohl(state);
	creatureData[creatureID].counterProtFromEvil = ntohl(protfromevil);
	creatureData[creatureID].counterDisease = ntohl(disease);
	creatureData[creatureID].counterPoison = ntohl(poison);
	creatureData[creatureID].counterBlindness = ntohl(blindness);
	creatureData[creatureID].counterInvisibility = ntohl(invisibility);
    }
}

void
SCClient::rcvAskForSpellsCast(Message *Msg)
{
    possibleLeftHandSpells.clear();
    possibleRightHandSpells.clear();

    int numSpells, index = 0;
    memcpy(&numSpells, Msg->data+index, sizeof(numSpells));
    index += sizeof(numSpells);
    numSpells = ntohl(numSpells);

    int spell;
    for (int i=0; i<numSpells; i++) {
	memcpy(&spell, Msg->data+index, sizeof(int));
	index += sizeof(int);
	possibleLeftHandSpells.push_back(Spell(ntohl(spell)));
    }

    // right hand spells
    memcpy(&numSpells, Msg->data+index, sizeof(numSpells));
    index += sizeof(numSpells);
    numSpells = ntohl(numSpells);
    
    for (int i=0; i<numSpells; i++) {
	memcpy(&spell, Msg->data+index, sizeof(int));
	index += sizeof(int);
	possibleRightHandSpells.push_back(Spell(ntohl(spell)));
    }
}

void
SCClient::rcvAskMonsterDirections(Message *Msg)
{
    unsigned int numMonsters;
    int index = 0;

    memcpy(&numMonsters, Msg->data+index, sizeof(int));
    index += sizeof(int);
    numMonsters = ntohl(numMonsters);
    monstersToDirect.clear();

    int monsterID;

    for (unsigned int i=0; i<numMonsters; i++) {
	memcpy(&monsterID, Msg->data+index, sizeof(int));
	index += sizeof(int);
	monsterID = ntohl(monsterID);

	monstersToDirect.push_back(monsterID);
    }

    int numTargets;
    memcpy(&numTargets, Msg->data+index, sizeof(int));
    index += sizeof(int);
    numTargets = ntohl(numTargets);

    validTargets.clear();
    int targetID;
    for (int i=0; i<numTargets; i++) {
	memcpy(&targetID, Msg->data+index, sizeof(int));
	index += sizeof(int);
	targetID = ntohl(targetID);
	validTargets.push_back(targetID);
    }
}

void
SCClient::rcvNewMonsterInfo(Message *Msg)
{
    int monsterID;
    int monsterType;
    int ownerID;

    memcpy(&monsterID, Msg->data, sizeof(int));
    monsterID = ntohl(monsterID);

    memcpy(&monsterType, Msg->data+sizeof(int), sizeof(int));
    monsterType = ntohl(monsterType);

    memcpy(&ownerID, Msg->data+sizeof(int)*2, sizeof(int));
    ownerID = ntohl(ownerID);

    string name = Msg->data+4*sizeof(int);

    // Make sure owner exists
    assert(creatureData.find(ownerID)!=creatureData.end());
    
    CreatureState cs;
    cs.name = name;
    cs.hitPoints = 0;
    cs.state = 0;
    cs.owner = ownerID; // need to set owner
    cs.type = monsterType;

    assert(creatureData.find(monsterID)==creatureData.end());

    creatureData[monsterID] = cs;
}

void
SCClient::rcvMonsterAttackInfo(Message *Msg)
{
    MonsterAttackResult mar;
    int index = 0;
    
    memcpy(&mar.source, Msg->data+index, sizeof(int));
    mar.source = ntohl(mar.source);
    index += sizeof(int);

    if (mar.source!=0) {
	memcpy(&mar.target, Msg->data+index, sizeof(int));
	mar.target = ntohl(mar.target);
	index += sizeof(int);
	memcpy(&mar.damage, Msg->data+index, sizeof(int));
	mar.damage = ntohl(mar.damage);
	index += sizeof(int);
	mar.resultMessage = Msg->data+index+sizeof(int);

	monsterResults.push_back(mar);
    } else {
	rcvdAllMonsterResults = true;
    }
}

void
SCClient::rcvEndGame(Message *Msg)
{
    int numWinners;
    int index = 0;
    int winner;
    gameWinners.clear();
    memcpy(&numWinners, Msg->data, sizeof(int));
    index += sizeof(int);
    numWinners = ntohl(numWinners);
    assert(numWinners>0);
    for (int i=0; i<numWinners; i++) {
	memcpy(&winner, Msg->data+index, sizeof(int));
	index += sizeof(int);
	winner = ntohl(winner);
	gameWinners.push_back(winner);
    }
}

void
SCClient::rcvEventMessage(Message *Msg)
{
    int index = 0;
    memcpy(&lastEventMessage.EventType, Msg->data, sizeof(int));
    index += sizeof(int);
    lastEventMessage.EventType = ntohl(lastEventMessage.EventType);
    
    memcpy(&lastEventMessage.MessageSource, Msg->data+index, sizeof(int));
    index += sizeof(int);
    lastEventMessage.MessageSource = ntohl(lastEventMessage.MessageSource);
    
    memcpy(&lastEventMessage.Source, Msg->data+index, sizeof(int));
    index += sizeof(int);
    lastEventMessage.Source = ntohl(lastEventMessage.Source);

    memcpy(&lastEventMessage.Target, Msg->data+index, sizeof(int));
    index += sizeof(int);
    lastEventMessage.Target = ntohl(lastEventMessage.Target);

    memcpy(&lastEventMessage.Misc, Msg->data+index, sizeof(int));
    index += sizeof(int);
    lastEventMessage.Misc = ntohl(lastEventMessage.Misc);

    // Skip size
    index += sizeof(int);
    lastEventMessage.MessageString = Msg->data+index;
}

void
SCClient::rcvCharmPersonCtrlHand(Message *Msg)
{
    memcpy(&charmTargetID, Msg->data, sizeof(int));
    charmTargetID = ntohl(charmTargetID);
}

void
SCClient::rcvParalysisCtrlHand(Message *Msg)
{
    memcpy(&paralysisTargetID, Msg->data, sizeof(int));
    paralysisTargetID = ntohl(paralysisTargetID);
}

void
SCClient::rcvCharmPersonCtrlGesture(Message *Msg)
{
    memcpy(&charmTargetID, Msg->data, sizeof(int));
    charmTargetID = ntohl(charmTargetID);
    memcpy(&charmedHand, Msg->data+sizeof(int), sizeof(int));
    charmedHand = ntohl(charmedHand);
}
